#include <math.h>
#include <cutil.h>
#include "solve.h"
#include "myutils.h"

template <class T> void array_swap_elements(T *a, int i, int j)
{
	T temp = a[i];
	a[i] = a[j];
	a[j] = temp;
}

//This function solves a set of linear equations of the form Ax = B using
//a standard Guassian elimination algorithm
void solve_serial(sparse_matrix A, float x[], const float xB[], float y[], const float yB[], const unsigned int array_dim)
{
	//Sanity check
	MYASSERT(array_dim == A.get_num_rows());

	//I'm going to need to muck around with the B matrix so copy it
	float * my_xB = new float[array_dim];
	float * my_yB = new float[array_dim];
	for (unsigned int c = 0; c < array_dim; c++)
	{
		my_xB[c] = xB[c];
		my_yB[c] = yB[c];
	}

	unsigned int htimer;
	cutCreateTimer (&htimer);
	cutStartTimer(htimer);
	for (unsigned int pivot_row = 0; pivot_row < A.get_num_rows(); pivot_row++)
	{
		unsigned int pivot_col = pivot_row;
		if (feq(A.get_entry(pivot_row, pivot_col), 0.0))
		{
			//Better find a new goddamn pivot
			unsigned int swap_row;
			for (swap_row = pivot_row + 1; swap_row < A.get_num_rows(); swap_row++)
			{
				if (!feq(A.get_entry(swap_row, pivot_col), 0.0))
				{
					A.swap(swap_row,pivot_row);
					array_swap_elements(my_xB, swap_row, pivot_row);
					array_swap_elements(my_yB, swap_row, pivot_row);
					break;
				}
			}
			MYASSERT(swap_row < A.get_num_rows());
		}
		//Clear out the pivot row
		float pivot_value = A.get_entry(pivot_row, pivot_col);
		for (unsigned int col = pivot_col; col < A.get_num_cols(); col++)
		{
			A.set_entry(pivot_row, col, A.get_entry(pivot_row, col) / pivot_value);
		}
		my_xB[pivot_row] /= pivot_value;
		my_yB[pivot_row] /= pivot_value;

		//Minus off all rows below it now
		for (unsigned int row = pivot_row + 1; row < A.get_num_rows(); row++)
		{
			pivot_value = A.get_entry(row, pivot_col);
			if (feq(pivot_value, 0.0))
				continue;

			for (unsigned int col = pivot_col; col <  A.get_num_cols(); col++)
			{
				A.set_entry(row, col, A.get_entry(row, col) - pivot_value * A.get_entry(pivot_row, col));
			}
			my_xB[row] -= pivot_value * my_xB[pivot_row];
			my_yB[row] -= pivot_value * my_yB[pivot_row];
		}
	}
	cutStopTimer(htimer);
	double time1 = cutGetTimerValue(htimer);
	printf("Upper triangular reduction time: %E\n", time1);
	cutResetTimer(htimer);
	cutStartTimer(htimer);

	//A.debug_print("in solve");
	//printf("myB %f %f %f%f\n", my_B[0], my_B[1],my_B[2], my_B[3]);
	
	//Matrix should now be upper triangular. Great success!
	//Now go back up from the bottom and solve for each variable
	for (signed int row = array_dim - 1; row >= 0; row--)
	{
		x[row] = my_xB[row];
		y[row] = my_yB[row];

		//printf("ser X[%d] = %f\n", row, x[row]);
		//printf("ser Y[%d] = %f\n", row, y[row]);

		for (unsigned int c = row + 1; c < array_dim; c++)
		{
			x[row] -= x[c] * A.get_entry(row, c);
			y[row] -= y[c] * A.get_entry(row, c);
		}
	}
	cutStopTimer(htimer);
	double time2 = cutGetTimerValue(htimer);
	printf("Back substitution time: %E\n", time2);

	printf("Total time: %E\n", time1 + time2);
	
	//Aaaaaannd we're done
	delete my_xB;
	delete my_yB;
}

//This function solves a set of linear equations of the form Ax = B using
//a standard Guassian elimination algorithm
void solve_cuda(sparse_matrix A, float x[], const float xB[], float y[], const float yB[], const unsigned int array_dim)
{
	//Sanity check
	MYASSERT(array_dim == A.get_num_rows());
	MYASSERT(array_dim < (1 << 16));

	// Create augmented matrix [A | xB | yB]
	unsigned int array_height = array_dim;
	unsigned int array_width = array_dim + 2;
	float *Maug = new float[array_height * array_width];
	for (unsigned int row = 0; row < array_dim; row++)
	{
		for (unsigned int col = 0; col < array_dim; col++)
		{
			Maug[row * array_width + col] = A.get_entry(row, col);
		}
		Maug[row * array_width + array_dim] = xB[row];
		Maug[row * array_width + array_dim + 1] = yB[row];
	}

	// Copy augment matrix Maug to device memory
	float *d_Maug;
	CUDA_SAFE_CALL(cudaMalloc((void **) &d_Maug, sizeof(float) * array_height * array_width));
	CUDA_SAFE_CALL(cudaMemcpy(d_Maug, Maug, sizeof(float) * array_height * array_width, cudaMemcpyHostToDevice));

	unsigned int *d_swap_row;
	CUDA_SAFE_CALL(cudaMalloc((void **) &d_swap_row, sizeof(unsigned int)));
	cudaThreadSynchronize();


	dim3 GridDims, BlockDims;
	unsigned int htimer;
	cutCreateTimer (&htimer);
	cutStartTimer(htimer);

	//dump(d_A, d_xB, d_yB, array_dim);

	unsigned int num_jobs;
	
	/*
	Perform upper-triangular reduction.
	This algorithm is slightly modified from a traditional gaussian elimination in that
	when the pivot row is normalized, all columns below it are also simultaneously normalized.
	This is so that these rows can be subtracted by the pivot row directly rather than having to
	be divided by the row's first non-zero element (which is essentially normalization anyway)
	*/
	for (unsigned int pivot_row = 0; pivot_row < array_dim; pivot_row++)
	{
		// Swap pivot row if with other row with non-zero lead column
		// and normalize the new pivot row to have leading 1
		/*
		DimCalc(1, GridDims, BlockDims);
		cudaThreadSynchronize();
		FindSwapRow <<< GridDims, BlockDims >>> (d_Maug, d_swap_row, pivot_row, array_width, array_height);

		*/
		num_jobs = array_width - pivot_row;
		DimCalc(num_jobs, GridDims, BlockDims);
		SwapAndNormalizePivotRow <<< GridDims, BlockDims>>> (d_Maug, d_swap_row, pivot_row, array_width, num_jobs);

		//dump(d_A, d_xB, d_yB, array_dim);



		// eliminate lead zeros all row below pivot row
		num_jobs = (array_height - pivot_row - 1) * (array_width - pivot_row);
		DimCalc(num_jobs, GridDims, BlockDims);
		cudaThreadSynchronize();
		EliminateRows <<< GridDims, BlockDims >>> (d_Maug, pivot_row, array_height, array_width, num_jobs);
		
		//dump(d_A, d_xB, d_yB, array_dim);

	}
	//dump_dev_array(d_Maug, array_height, array_width);

	cudaThreadSynchronize();
	cutStopTimer(htimer);
	double time1 = cutGetTimerValue(htimer);
	printf("Upper triangular reduction time: %E\n", time1);
	cutResetTimer(htimer);
	cutStartTimer(htimer);

	/*
	Matrix should now be upper triangular. Great success!
	Now go back up from the bottom and solve for each variable
	In a simple upper-triangular matrix like so
	| 1 A1 A2 | |x1|   |B1|
	| 0 1  A3 | |x2| = |B2|
	| 0 1  1  | |x3| = |B3|

	The equation to solve for x1 in each row is as follows
	x1 = B1 - (A1 * x1 + A2 * x2)
	This involves 3 operations: Performing a per-element multiplication of the first row of the A matrix
	and the X matrix, adding the results up, and finally subtracting it from the B value.
	*/
	for (signed int value_row = array_dim - 1; value_row > 0; value_row--)
	{
		num_jobs = value_row;
		DimCalc(num_jobs, GridDims, BlockDims);
		cudaThreadSynchronize();
		SubtractValue <<< GridDims, BlockDims >>> (d_Maug, value_row, array_height, array_width, array_dim, num_jobs);
	}

	cutStopTimer(htimer);
	double time2 = cutGetTimerValue(htimer);
	printf("Back substitution time: %E\n", time2);
	printf("Total time: %E\n", time1 + time2);

	//Aaaaaannd we're done
	CUDA_SAFE_CALL(cudaMemcpy((void *) Maug, d_Maug, sizeof(float) * array_height * array_width, cudaMemcpyDeviceToHost));
	cudaThreadSynchronize();
	for (unsigned int row = 0; row < array_dim; row++)
	{
		x[row] = Maug[row * array_width + array_dim];
		y[row] = Maug[row * array_width + array_dim + 1];
	}
}

/*
Handy utility function that helps calculate pesky dimensions
In:		jobs - Number of threads to lanch
Out:	GridDims - Calculated grid dimensions
		BlockDims - Calculate block dimensions
*/
void DimCalc(unsigned int jobs, dim3 & GridDims, dim3 & BlockDims)
{
	unsigned int const MaxthreadDim = 512;
	unsigned int const MaxblockDim = 65536;
	
	// Saturate threadDim first
	if (jobs < MaxthreadDim) {BlockDims.x = jobs;} else {BlockDims.x = MaxthreadDim;}

	// Saturate blockDim.x next
	for (GridDims.x = 1; GridDims.x * BlockDims.x < jobs && GridDims.x < MaxblockDim; GridDims.x++);

	// Saturate blockDim.y last
	for (GridDims.y = 1; GridDims.y * GridDims.x * BlockDims.x < jobs && GridDims.y < MaxblockDim; GridDims.y++);

}

__global__ void FindSwapRow(float * d_in, unsigned int * d_swap_row, unsigned int pivot_row, unsigned int array_width, unsigned int array_height)
{
	unsigned int col = pivot_row;
	if (!feq(d_in[pivot_row * array_width + col], 0.0))
	{
		*d_swap_row = 0;
		return;
	}
	else
	{
		//Better find a new pivot
		for (unsigned int swap_row = pivot_row + 1; swap_row < array_height; swap_row++)
		{
			if (!feq(d_in[swap_row * array_width + col], 0.0))
			{
				*d_swap_row = swap_row;
				return;
			}
		}
	}
}

__global__ void SwapAndNormalizePivotRow(float * d_in, unsigned int * d_swap_row, unsigned int pivot_row, unsigned int array_width, unsigned int num_jobs)
{
	unsigned int i = (blockIdx.y * gridDim.x + blockIdx.x) * blockDim.x + threadIdx.x;
	if (i < num_jobs)
	{
		unsigned int col = pivot_row + i;

		//unsigned int swap_row = *d_swap_row;
		//if (swap_row == 0)
		//{
			d_in[pivot_row * array_width + col] = d_in[pivot_row * array_width + col] / d_in[pivot_row * array_width + pivot_row];
		//}
		/*
		else
		{
			float temp = d_in[pivot_row * array_width + col];
			d_in[pivot_row * array_width + col] = d_in[swap_row * array_width + col] / d_in[swap_row * array_width + pivot_row];
			d_in[swap_row * array_width + col] = temp;

		}
		*/
	}
}

/*
__global__ void SwapAndNormalizePivotRow(float *d_Maug, unsigned int pivot_row, const unsigned int array_height, const unsigned int array_width)
{
	unsigned int pivot_col = pivot_row;
	if (feq(d_Maug[pivot_row * array_width + pivot_col], 0.0))
	{
		//Better find a new pivot
		for (unsigned int swap_row = pivot_row + 1; swap_row < array_height; swap_row++)
		{
			float pivot_value = d_Maug[swap_row * array_width + pivot_col];
			if (!feq(pivot_value, 0.0))
			{
				unsigned int col = pivot_col + (blockIdx.y * gridDim.x + blockIdx.x) * blockDim.x + threadIdx.x;
				float temp = d_Maug[pivot_row * array_width + col];
				d_Maug[pivot_row * array_width + col] = d_Maug[swap_row * array_width + col] / pivot_value;
				d_Maug[swap_row * array_width + col] = temp;

				return;
			}
		}
	}
}
*/

__global__ void EliminateRows(float *d_Maug, unsigned int pivot_row, unsigned int array_height, unsigned int array_width, unsigned int num_jobs)
{
	unsigned int i = (blockIdx.y * gridDim.x + blockIdx.x) * blockDim.x + threadIdx.x;
	if (i < num_jobs)
	{
		unsigned int starting_row = pivot_row + 1;
		unsigned int starting_col = pivot_row;
		unsigned int row = starting_row +  (i / (array_width - pivot_row));
		unsigned int col = starting_col + (i % (array_width - pivot_row));

		float leading_value = d_Maug[row * array_width + starting_col];
		if (!feq(leading_value, 0.0))
		{
			d_Maug[row * array_width + col] -= (leading_value * d_Maug[pivot_row * array_width + col]);
		}
	}
}

__global__ void SubtractValue(float *d_Maug, unsigned int value_row, unsigned int array_height, unsigned int array_width, unsigned int array_dim, unsigned int num_jobs)
{
	unsigned int i = (blockIdx.y * gridDim.x + blockIdx.x) * blockDim.x + threadIdx.x;
	if (i < num_jobs)
	{
		unsigned int row = i;

		// xB
		d_Maug[row * array_width + array_dim] -= d_Maug[row * array_width + value_row] * d_Maug[value_row * array_width + array_dim];

		// yB
		d_Maug[row * array_width + array_dim + 1] -= d_Maug[row * array_width + value_row] * d_Maug[value_row * array_width + array_dim + 1];
	}
}

// Below are a bunch of dumper functions useful for debugging
void dump_dev_array(unsigned int *d_A, unsigned int array_height, unsigned int array_width)
{
	unsigned int *A = (unsigned int *) malloc(sizeof(float) * array_height * array_width);
	CUDA_SAFE_CALL(cudaMemcpy((void *) A, d_A, sizeof(float) * array_height * array_width, cudaMemcpyDeviceToHost));
	cudaThreadSynchronize();

	printf("A = \n ");
	for (unsigned int i = 0; i < array_height && i < 12; i++)
	{
		for (unsigned int j = 0; j < array_width && j < 12; j++)
		{
			printf("%3d ", A[i * array_width + j]);
		}
		printf("\n");
	}
}

void dump_dev_array(float *d_A, unsigned int array_height, unsigned int array_width)
{
	float *A = (float *) malloc(sizeof(float) * array_height * array_width);
	CUDA_SAFE_CALL(cudaMemcpy((void *) A, d_A, sizeof(float) * array_height * array_width, cudaMemcpyDeviceToHost));
	cudaThreadSynchronize();

	printf("A = \n ");
	for (unsigned int i = 0; i < array_height && i < 12; i++)
	{
		for (unsigned int j = 0; j < array_width && j < 12; j++)
		{
			printf("%5.1f ", A[i * array_width + j]);
		}
		printf("\n");
	}
}

void dump(float * d_A, float * d_xB, float * d_yB, unsigned int size)
{
	printf("A = \n ");
	for (unsigned int i = 0; i < size; i++)
	{
		for (unsigned int j = 0; j < size; j++)
		{
			printf("%f ", d_A[i * size + j]);
		}
		printf("\n");
	}
	printf("\nBx = ");
	for (unsigned int c = 0; c < size; c++)
	{
		printf("%f ", d_xB[c]);
	}
	printf("\nBy = ");
	for (unsigned int c = 0; c < size; c++)
	{
		printf("%f ", d_yB[c]);
	}
	printf("\n\n");
}

void DumpReductionBuffer(float * d_Buf1, float * d_Buf2, unsigned int size)
{
	printf("X = \n");
	for (unsigned int c = 0; c < size; c++)
	{
		printf("%f ", d_Buf1[c]);
	}
	printf("\nY = \n");
	for (unsigned int c = 0; c < size; c++)
	{
		printf("%f ", d_Buf2[c]);
	}
	printf("\n\n");
}
